"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeBadDefinitionTypeMessage = exports.parseObject = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const objectLiteral_js_1 = require("./objectLiteral.js");
const tuple_js_1 = require("./tuple.js");
const parseObject = (def, ctx) => {
    const objectKind = (0, util_1.objectKindOf)(def);
    switch (objectKind) {
        case undefined:
            if ((0, schema_1.hasArkKind)(def, "root"))
                return def;
            return (0, objectLiteral_js_1.parseObjectLiteral)(def, ctx);
        case "Array":
            return (0, tuple_js_1.parseTuple)(def, ctx);
        case "RegExp":
            return ctx.$.node("intersection", {
                domain: "string",
                regex: def
            }, { prereduced: true });
        case "Function": {
            const resolvedDef = (0, util_1.isThunk)(def) ? def() : def;
            if ((0, schema_1.hasArkKind)(resolvedDef, "root"))
                return resolvedDef;
            return (0, util_1.throwParseError)((0, exports.writeBadDefinitionTypeMessage)("Function"));
        }
        default:
            return (0, util_1.throwParseError)((0, exports.writeBadDefinitionTypeMessage)(objectKind ?? (0, util_1.printable)(def)));
    }
};
exports.parseObject = parseObject;
const writeBadDefinitionTypeMessage = (actual) => `Type definitions must be strings or objects (was ${actual})`;
exports.writeBadDefinitionTypeMessage = writeBadDefinitionTypeMessage;
