"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.tryParseDate = exports.writeInvalidDateMessage = exports.extractDateLiteralSource = exports.isValidDate = exports.isDateLiteral = void 0;
const util_1 = require("@arktype/util");
const isDateLiteral = (value) => typeof value === "string" &&
    value[0] === "d" &&
    (value[1] === "'" || value[1] === '"') &&
    value.at(-1) === value[1];
exports.isDateLiteral = isDateLiteral;
const isValidDate = (d) => d.toString() !== "Invalid Date";
exports.isValidDate = isValidDate;
const extractDateLiteralSource = (literal) => literal.slice(2, -1);
exports.extractDateLiteralSource = extractDateLiteralSource;
const writeInvalidDateMessage = (source) => `'${source}' could not be parsed by the Date constructor`;
exports.writeInvalidDateMessage = writeInvalidDateMessage;
const tryParseDate = (source, errorOnFail) => maybeParseDate(source, errorOnFail);
exports.tryParseDate = tryParseDate;
const maybeParseDate = (source, errorOnFail) => {
    const stringParsedDate = new Date(source);
    if ((0, exports.isValidDate)(stringParsedDate))
        return stringParsedDate;
    const epochMillis = (0, util_1.tryParseNumber)(source);
    if (epochMillis !== undefined) {
        const numberParsedDate = new Date(epochMillis);
        if ((0, exports.isValidDate)(numberParsedDate))
            return numberParsedDate;
    }
    return errorOnFail ?
        (0, util_1.throwParseError)(errorOnFail === true ? (0, exports.writeInvalidDateMessage)(source) : errorOnFail)
        : undefined;
};
