"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeUnterminatedEnclosedMessage = exports.untilLookaheadIsClosing = exports.enclosingTokens = exports.enclosingChar = exports.enclosingQuote = exports.parseEnclosed = void 0;
const util_1 = require("@arktype/util");
const date_js_1 = require("./date.js");
const parseEnclosed = (s, enclosing) => {
    const enclosed = s.scanner.shiftUntil(exports.untilLookaheadIsClosing[exports.enclosingTokens[enclosing]]);
    if (s.scanner.lookahead === "")
        return s.error((0, exports.writeUnterminatedEnclosedMessage)(enclosed, enclosing));
    // Shift the scanner one additional time for the second enclosing token
    s.scanner.shift();
    if (enclosing === "/") {
        // fail parsing if the regex is invalid
        new RegExp(enclosed);
        s.root = s.ctx.$.node("intersection", {
            domain: "string",
            regex: enclosed
        }, { prereduced: true });
    }
    else if ((0, util_1.isKeyOf)(enclosing, exports.enclosingQuote))
        s.root = s.ctx.$.node("unit", { unit: enclosed });
    else {
        const date = (0, date_js_1.tryParseDate)(enclosed, (0, date_js_1.writeInvalidDateMessage)(enclosed));
        s.root = s.ctx.$.node("unit", { unit: date, description: enclosed });
    }
};
exports.parseEnclosed = parseEnclosed;
exports.enclosingQuote = {
    "'": 1,
    '"': 1
};
exports.enclosingChar = {
    "/": 1,
    "'": 1,
    '"': 1
};
exports.enclosingTokens = {
    "d'": "'",
    'd"': '"',
    "'": "'",
    '"': '"',
    "/": "/"
};
exports.untilLookaheadIsClosing = {
    "'": scanner => scanner.lookahead === `'`,
    '"': scanner => scanner.lookahead === `"`,
    "/": scanner => scanner.lookahead === `/`
};
const enclosingCharDescriptions = {
    '"': "double-quote",
    "'": "single-quote",
    "/": "forward slash"
};
const writeUnterminatedEnclosedMessage = (fragment, enclosingStart) => `${enclosingStart}${fragment} requires a closing ${enclosingCharDescriptions[exports.enclosingTokens[enclosingStart]]}`;
exports.writeUnterminatedEnclosedMessage = writeUnterminatedEnclosedMessage;
