"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeExpressionExpectedMessage = exports.writeMissingRightOperandMessage = exports.writeMissingOperandMessage = exports.parseGenericInstantiation = exports.parseUnenclosed = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const validate_js_1 = require("../../../semantic/validate.js");
const genericArgs_js_1 = require("./genericArgs.js");
const parseUnenclosed = (s) => {
    const token = s.scanner.shiftUntilNextTerminator();
    if (token === "keyof")
        s.addPrefix("keyof");
    else
        s.root = unenclosedToNode(s, token);
};
exports.parseUnenclosed = parseUnenclosed;
const parseGenericInstantiation = (name, g, s) => {
    s.scanner.shiftUntilNonWhitespace();
    const lookahead = s.scanner.shift();
    if (lookahead !== "<")
        return s.error((0, genericArgs_js_1.writeInvalidGenericArgsMessage)(name, g.params, []));
    const parsedArgs = (0, genericArgs_js_1.parseGenericArgs)(name, g.params, s);
    const remainingChars = parsedArgs.unscanned.length;
    // set the scanner position to where the args scanner left off
    s.scanner.jumpToIndex(remainingChars === 0 ? s.scanner.length : -remainingChars);
    return g(...parsedArgs.result);
};
exports.parseGenericInstantiation = parseGenericInstantiation;
const unenclosedToNode = (s, token) => maybeParseReference(s, token) ??
    maybeParseUnenclosedLiteral(s, token) ??
    s.error(token === "" ? (0, exports.writeMissingOperandMessage)(s)
        : token[0] === "#" ?
            (0, validate_js_1.writePrefixedPrivateReferenceMessage)(token)
            : (0, schema_1.writeUnresolvableMessage)(token));
const maybeParseReference = (s, token) => {
    if (s.ctx.args?.[token])
        return s.ctx.args[token].raw;
    const resolution = s.ctx.$.maybeResolve(token);
    if (resolution instanceof schema_1.BaseRoot)
        return resolution;
    if (resolution === undefined)
        return;
    if ((0, schema_1.hasArkKind)(resolution, "generic"))
        return (0, exports.parseGenericInstantiation)(token, resolution, s);
    return (0, util_1.throwParseError)(`Unexpected resolution ${(0, util_1.printable)(resolution)}`);
};
const maybeParseUnenclosedLiteral = (s, token) => {
    const maybeNumber = (0, util_1.tryParseNumber)(token, { strict: true });
    if (maybeNumber !== undefined)
        return s.ctx.$.node("unit", { unit: maybeNumber });
    const maybeBigint = (0, util_1.tryParseWellFormedBigint)(token);
    if (maybeBigint !== undefined)
        return s.ctx.$.node("unit", { unit: maybeBigint });
};
const writeMissingOperandMessage = (s) => {
    const operator = s.previousOperator();
    return operator ?
        (0, exports.writeMissingRightOperandMessage)(operator, s.scanner.unscanned)
        : (0, exports.writeExpressionExpectedMessage)(s.scanner.unscanned);
};
exports.writeMissingOperandMessage = writeMissingOperandMessage;
const writeMissingRightOperandMessage = (token, unscanned = "") => `Token '${token}' requires a right operand${unscanned ? ` before '${unscanned}'` : ""}`;
exports.writeMissingRightOperandMessage = writeMissingRightOperandMessage;
const writeExpressionExpectedMessage = (unscanned) => `Expected an expression${unscanned ? ` before '${unscanned}'` : ""}`;
exports.writeExpressionExpectedMessage = writeExpressionExpectedMessage;
