"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeInvalidConstructorMessage = exports.parseNarrowTuple = exports.writeMalformedFunctionalExpressionMessage = exports.parseMorphTuple = exports.parseKeyOfTuple = exports.spreadOptionalMessage = exports.optionalPostVariadicMessage = exports.requiredPostOptionalMessage = exports.multipleVariadicMesage = exports.writeNonArraySpreadMessage = exports.parseTupleLiteral = exports.parseTuple = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const unenclosed_js_1 = require("./string/shift/operand/unenclosed.js");
const parseTuple = (def, ctx) => maybeParseTupleExpression(def, ctx) ?? (0, exports.parseTupleLiteral)(def, ctx);
exports.parseTuple = parseTuple;
const parseTupleLiteral = (def, ctx) => {
    let sequences = [{}];
    let i = 0;
    while (i < def.length) {
        let spread = false;
        let optional = false;
        if (def[i] === "..." && i < def.length - 1) {
            spread = true;
            i++;
        }
        const element = ctx.$.parse(def[i], ctx);
        i++;
        if (def[i] === "?") {
            if (spread)
                return (0, util_1.throwParseError)(exports.spreadOptionalMessage);
            optional = true;
            i++;
        }
        if (spread) {
            if (!element.extends(schema_1.jsObjects.Array))
                return (0, util_1.throwParseError)((0, exports.writeNonArraySpreadMessage)(element.expression));
            // a spread must be distributed over branches e.g.:
            // def: [string, ...(number[] | [true, false])]
            // nodes: [string, ...number[]] | [string, true, false]
            sequences = sequences.flatMap(base => 
            // since appendElement mutates base, we have to shallow-ish clone it for each branch
            element.branches.map(branch => appendSpreadBranch((0, schema_1.makeRootAndArrayPropertiesMutable)(base), branch)));
        }
        else {
            sequences = sequences.map(base => appendElement(base, optional ? "optional" : "required", element));
        }
    }
    return ctx.$.raw.schema(sequences.map(sequence => ({
        proto: Array,
        sequence
    })));
};
exports.parseTupleLiteral = parseTupleLiteral;
const appendElement = (base, kind, element) => {
    switch (kind) {
        case "required":
            if (base.optionals)
                // e.g. [string?, number]
                return (0, util_1.throwParseError)(exports.requiredPostOptionalMessage);
            if (base.variadic) {
                // e.g. [...string[], number]
                base.postfix = (0, util_1.append)(base.postfix, element);
            }
            else {
                // e.g. [string, number]
                base.prefix = (0, util_1.append)(base.prefix, element);
            }
            return base;
        case "optional":
            if (base.variadic)
                // e.g. [...string[], number?]
                return (0, util_1.throwParseError)(exports.optionalPostVariadicMessage);
            // e.g. [string, number?]
            base.optionals = (0, util_1.append)(base.optionals, element);
            return base;
        case "variadic":
            // e.g. [...string[], number, ...string[]]
            if (base.postfix)
                (0, util_1.throwParseError)(exports.multipleVariadicMesage);
            if (base.variadic) {
                if (!base.variadic.equals(element)) {
                    // e.g. [...string[], ...number[]]
                    (0, util_1.throwParseError)(exports.multipleVariadicMesage);
                }
                // e.g. [...string[], ...string[]]
                // do nothing, second spread doesn't change the type
            }
            else {
                // e.g. [string, ...number[]]
                base.variadic = element.raw;
            }
            return base;
    }
};
const appendSpreadBranch = (base, branch) => {
    const spread = branch.firstReferenceOfKind("sequence");
    if (!spread) {
        // the only array with no sequence reference is unknown[]
        return appendElement(base, "variadic", schema_1.tsKeywords.unknown);
    }
    spread.prefix.forEach(node => appendElement(base, "required", node));
    spread.optionals.forEach(node => appendElement(base, "optional", node));
    spread.variadic && appendElement(base, "variadic", spread.variadic);
    spread.postfix.forEach(node => appendElement(base, "required", node));
    return base;
};
const maybeParseTupleExpression = (def, ctx) => {
    const tupleExpressionResult = isIndexZeroExpression(def) ? prefixParsers[def[0]](def, ctx)
        : isIndexOneExpression(def) ? indexOneParsers[def[1]](def, ctx)
            : undefined;
    return tupleExpressionResult;
    // TODO: remove
    // return tupleExpressionResult.isNever()
    // 	? throwParseError(
    // 			writeUnsatisfiableExpressionError(
    // 				def
    // 					.map((def) =>
    // 						typeof def === "string" ? def : printable(def)
    // 					)
    // 					.join(" ")
    // 			)
    // 		)
    // 	: tupleExpressionResult
};
const writeNonArraySpreadMessage = (operand) => `Spread element must be an array (was ${operand})`;
exports.writeNonArraySpreadMessage = writeNonArraySpreadMessage;
exports.multipleVariadicMesage = "A tuple may have at most one variadic element";
exports.requiredPostOptionalMessage = "A required element may not follow an optional element";
exports.optionalPostVariadicMessage = "An optional element may not follow a variadic element";
exports.spreadOptionalMessage = "A spread element cannot be optional";
const parseKeyOfTuple = (def, ctx) => ctx.$.parse(def[1], ctx).keyof();
exports.parseKeyOfTuple = parseKeyOfTuple;
const parseBranchTuple = (def, ctx) => {
    if (def[2] === undefined)
        return (0, util_1.throwParseError)((0, unenclosed_js_1.writeMissingRightOperandMessage)(def[1], ""));
    const l = ctx.$.parse(def[0], ctx);
    const r = ctx.$.parse(def[2], ctx);
    return def[1] === "&" ? l.and(r) : l.or(r);
};
const parseArrayTuple = (def, ctx) => ctx.$.parse(def[0], ctx).array();
const isIndexOneExpression = (def) => indexOneParsers[def[1]] !== undefined;
const parseMorphTuple = (def, ctx) => {
    if (typeof def[2] !== "function") {
        return (0, util_1.throwParseError)((0, exports.writeMalformedFunctionalExpressionMessage)("=>", def[2]));
    }
    // TODO: nested morphs?
    return ctx.$.parse(def[0], ctx).pipe(def[2]);
};
exports.parseMorphTuple = parseMorphTuple;
const writeMalformedFunctionalExpressionMessage = (operator, value) => `${operator === ":" ? "Narrow" : "Morph"} expression requires a function following '${operator}' (was ${typeof value})`;
exports.writeMalformedFunctionalExpressionMessage = writeMalformedFunctionalExpressionMessage;
const parseNarrowTuple = (def, ctx) => {
    if (typeof def[2] !== "function") {
        return (0, util_1.throwParseError)((0, exports.writeMalformedFunctionalExpressionMessage)(":", def[2]));
    }
    return ctx.$.parse(def[0], ctx).constrain("predicate", def[2]);
};
exports.parseNarrowTuple = parseNarrowTuple;
const parseAttributeTuple = (def, ctx) => ctx.$.parse(def[0], ctx).configureShallowDescendants(def[2]);
const indexOneParsers = {
    "|": parseBranchTuple,
    "&": parseBranchTuple,
    "[]": parseArrayTuple,
    ":": exports.parseNarrowTuple,
    "=>": exports.parseMorphTuple,
    "@": parseAttributeTuple
};
const prefixParsers = {
    keyof: exports.parseKeyOfTuple,
    instanceof: (def, ctx) => {
        if (typeof def[1] !== "function") {
            return (0, util_1.throwParseError)((0, exports.writeInvalidConstructorMessage)((0, util_1.objectKindOrDomainOf)(def[1])));
        }
        const branches = def
            .slice(1)
            .map(ctor => typeof ctor === "function" ?
            ctx.$.node("proto", { proto: ctor })
            : (0, util_1.throwParseError)((0, exports.writeInvalidConstructorMessage)((0, util_1.objectKindOrDomainOf)(ctor))));
        return branches.length === 1 ?
            branches[0]
            : ctx.$.node("union", { branches });
    },
    "===": (def, ctx) => ctx.$.units(def.slice(1))
};
const isIndexZeroExpression = (def) => prefixParsers[def[0]] !== undefined;
const writeInvalidConstructorMessage = (actual) => `Expected a constructor following 'instanceof' operator (was ${actual})`;
exports.writeInvalidConstructorMessage = writeInvalidConstructorMessage;
