"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseScopeKey = exports.writeShallowCycleErrorMessage = exports.RawScope = exports.scope = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const generic_js_1 = require("./generic.js");
const match_js_1 = require("./match.js");
const definition_js_1 = require("./parser/definition.js");
const generic_js_2 = require("./parser/generic.js");
const dynamic_js_1 = require("./parser/string/reduce/dynamic.js");
const string_js_1 = require("./parser/string/string.js");
const type_js_1 = require("./type.js");
exports.scope = ((def, config = {}) => new RawScope(def, config));
class RawScope extends schema_1.RawRootScope {
    parseCache = {};
    constructor(def, config) {
        const aliases = {};
        for (const k in def) {
            const parsedKey = (0, exports.parseScopeKey)(k);
            aliases[parsedKey.name] =
                parsedKey.params.length ?
                    // TODO: this
                    new generic_js_1.Generic(parsedKey.params, def[k], {})
                    : def[k];
        }
        super(aliases, config);
    }
    type = new type_js_1.RawTypeParser(this);
    match = (0, match_js_1.createMatchParser)(this);
    declare = (() => ({
        type: this.type
    })).bind(this);
    define = ((def) => def).bind(this);
    preparseRoot(def) {
        if ((0, util_1.isThunk)(def) && !(0, schema_1.hasArkKind)(def, "generic"))
            return def();
        return def;
    }
    parseRoot(def) {
        // args: { this: {} as RawRoot },
        return this.parse(def, {
            $: this,
            args: {}
            // type parsing can bypass nodes if it hits the cache,
            // so bind it directly (could be optimized)
        }).bindScope(this);
    }
    parse(def, ctx) {
        if (typeof def === "string") {
            if (ctx.args && Object.keys(ctx.args).every(k => !def.includes(k))) {
                // we can only rely on the cache if there are no contextual
                // resolutions like "this" or generic args
                return this.parseString(def, ctx);
            }
            if (!this.parseCache[def])
                this.parseCache[def] = this.parseString(def, ctx);
            return this.parseCache[def];
        }
        return (0, util_1.hasDomain)(def, "object") ?
            (0, definition_js_1.parseObject)(def, ctx)
            : (0, util_1.throwParseError)((0, definition_js_1.writeBadDefinitionTypeMessage)((0, util_1.domainOf)(def)));
    }
    parseString(def, ctx) {
        return (this.maybeResolveRoot(def) ??
            ((def.endsWith("[]") &&
                this.maybeResolveRoot(def.slice(0, -2))?.array()) ||
                (0, string_js_1.fullStringParse)(new dynamic_js_1.DynamicState(def, ctx))));
    }
}
exports.RawScope = RawScope;
const writeShallowCycleErrorMessage = (name, seen) => `Alias '${name}' has a shallow resolution cycle: ${[...seen, name].join(":")}`;
exports.writeShallowCycleErrorMessage = writeShallowCycleErrorMessage;
const parseScopeKey = (k) => {
    const firstParamIndex = k.indexOf("<");
    if (firstParamIndex === -1) {
        return {
            name: k,
            params: []
        };
    }
    if (k.at(-1) !== ">") {
        (0, util_1.throwParseError)(`'>' must be the last character of a generic declaration in a scope`);
    }
    return {
        name: k.slice(0, firstParamIndex),
        params: (0, generic_js_2.parseGenericParams)(k.slice(firstParamIndex + 1, -1))
    };
};
exports.parseScopeKey = parseScopeKey;
