"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Type = exports.RawTypeParser = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const generic_js_1 = require("./generic.js");
const generic_js_2 = require("./parser/generic.js");
const typeParserAttachments = Object.freeze({
    errors: schema_1.ArkErrors
});
class RawTypeParser extends util_1.Callable {
    constructor($) {
        super((...args) => {
            if (args.length === 1) {
                // treat as a simple definition
                return $.parseRoot(args[0]);
            }
            if (args.length === 2 &&
                typeof args[0] === "string" &&
                args[0][0] === "<" &&
                args[0].at(-1) === ">") {
                // if there are exactly two args, the first of which looks like <${string}>,
                // treat as a generic
                const params = (0, generic_js_2.parseGenericParams)(args[0].slice(1, -1));
                const def = args[1];
                // TODO: validateUninstantiatedGeneric, remove this cast
                return new generic_js_1.Generic(params, def, $);
            }
            // otherwise, treat as a tuple expression. technically, this also allows
            // non-expression tuple definitions to be parsed, but it's not a supported
            // part of the API as specified by the associated types
            return $.parseRoot(args);
        }, { bind: $, attach: typeParserAttachments });
    }
}
exports.RawTypeParser = RawTypeParser;
exports.Type = schema_1.BaseRoot;
